from numbers import Number
from nptyping import *
from typing import *
from numpy import array, set_printoptions, arange, zeros, interp, mean
from pandas import DataFrame
from copy import deepcopy, copy
from time import process_time
from numba import njit, jit
from numba.typed import List

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

import logging
import time
import builtins

set_printoptions(precision=4)
# from .utility_functions import *
from .esbmtk import esbmtkBase, Reservoir, Flux, Signal, Source, Sink
from .utility_functions import sort_by_type
from .solver import get_imass, get_frac, get_delta, get_flux_data, get_l_mass
from . import ureg, Q_

# from .connections import ConnnectionGroup


class Process(esbmtkBase):
    """This class defines template for process which acts on one or more
    reservoir flux combinations. To use it, you need to create an
    subclass which defines the actual process implementation in their
    call method. See 'PassiveFlux as example'

    """

    __slots__ = ("reservoir", "r", "flux", "r", "mo", "direction", "scale")

    def __init__(self, **kwargs: Dict[str, any]) -> None:
        """
        Create a new process object with a given process type and options
        """

        self.__defaultnames__()  # default kwargs names
        self.__initerrormessages__()  # default error messages
        self.bem.update({"rate": "a string"})
        self.bem.update({"scale": "Number or quantity"})
        self.__validateandregister__(kwargs)  # initialize keyword values

        self.__postinit__()  # do some housekeeping
        self.__register_name__()

    def __postinit__(self) -> None:
        """Do some housekeeping for the process class"""

        # legacy name aliases
        self.n: str = self.name  # display name of species
        if "reserevoir" in self.kwargs:
            self.r: Reservoir = self.reservoir
            self.m: Model = self.r.sp.mo
        else:
            self.m: Model = self.flux.mo  # the model handle

        self.mo: Model = self.m
        self.f: Flux = self.flux
        self.mo: Model = self.m

        # self.rm0: float = self.r.m[0]  # the initial reservoir mass

        if "reserevoir" in self.kwargs:
            if isinstance(self.r, Reservoir):
                self.direction: Dict[Flux, int] = self.r.lio[self.f]

        if "delta" in self.kwargs:
            self.delta = self.kwargs["delta"]
        else:
            self.delta = "None"

        self.__misc_init__()

    def __delayed_init__(self) -> None:
        """
        Initialize stuff which is only known after the entire model has been defined.
        This will be executed just before running the model. You need to add the following
        two lines somewhere in the init procedure (preferably by redefining __misc_init__)
        and redefine __delayed_init__ to actually execute any code below

        # this process requires delayed init
        self.mo.lto.append(self)

        """

        pass

    def __misc_init__(self) -> None:
        """This is just a place holder method which will be called by default
        in __post_init__() This can be overloaded to add additional
        code to the init procedure without the need to redefine
        init. This useful for processes which only define a call method.

        """

        pass

    def __defaultnames__(self) -> None:
        """Set up the default names and dicts for the process class. This
        allows us to extend these values without modifying the entire init process

        """

        from .connections import ConnectionGroup
        from esbmtk import Reservoir, ReservoirGroup, Flux, GasReservoir

        # provide a dict of known keywords and types
        self.lkk: Dict[str, any] = {
            "name": str,
            "reservoir": (Reservoir, Source, Sink, GasReservoir),
            "flux": Flux,
            "rate": (Number, np.float64),
            "delta": (Number, np.float64, str),
            "lt": Flux,
            "alpha": (Number, np.float64),
            "scale": (Number, np.float64),
            "ref_reservoirs": (Flux, Reservoir, GasReservoir, list, str),
            "register": (
                str,
                ConnectionGroup,
                Reservoir,
                ReservoirGroup,
                GasReservoir,
                Flux,
            ),
        }

        # provide a list of absolutely required keywords
        self.lrk: list[str] = ["name"]

        # list of default values if none provided
        self.lod: Dict[str, any] = {"scale": 1}

    def __register__(self, reservoir: Reservoir, flux: Flux) -> None:
        """Register the flux/reservoir pair we are acting upon, and register
        the process with the reservoir

        """

        # register the reservoir flux combination we are acting on
        self.f: Flux = flux
        self.r: Reservoir = reservoir
        # add this process to the list of processes acting on this reservoir
        reservoir.lop.append(self)
        flux.lop.append(self)
        # Add to model flux list
        reservoir.mo.lop.append(self)

    def show_figure(self, x, y) -> None:
        """Apply the current process to the vector x, and show the result as y.
        The resulting figure will be automatically saved.

        Example::
             process_name.show_figure(x,y)

        """
        pass


class GenericFunction(Process):
    """This Process class creates a GenericFunction instance which is
    typically used with the VirtualReservoir, and
    ExternalCode classes. This class is not user facing,
    please see the ExternalCode class docstring for the
    function template of a user provided function.

    see calc_carbonates in the carbonate chemistry for an example how
    to write a function for this class.

    """

    __slots__ = ("function", "input_data", "vr_data", "params")

    def __init__(self, **kwargs: Dict[str, any]) -> None:
        """
        Create a new process object with a given process type and options

        """

        from . import Reservoir_no_set

        self.__defaultnames__()  # default kwargs names

        # list of allowed keywords
        self.lkk: Dict[str, any] = {
            "name": str,
            "function": any,
            "input_data": (List, str),
            "vr_data": (List, str),
            "function_params": (List, str),
            "model": any,
        }

        # required arguments
        self.lrk: list = ["name", "input_data", "vr_data", "function_params", "model"]

        # list of default values if none provided
        self.lod: Dict[any, any] = {}

        self.__initerrormessages__()  # default error messages
        self.bem.update(
            {
                "function": "a function",
                "input_data": "list of one or more numpy arrays",
                "vr_data": "list of one or more numpy arrays",
                "function_params": "a list of float values",
            }
        )
        self.__validateandregister__(kwargs)  # initialize keyword values
        self.mo = self.model

        if not callable(self.function):
            raise ValueError("function must be defined before it can be used here")

        self.__postinit__()  # do some housekeeping

        if self.mo.register == "local" and self.register == "None":
            self.register = self.mo

        self.__register_name__()  #

    def __call__(self, i: int) -> None:
        """Here we execute the user supplied function
        Where i = index of the current timestep

        """

        self.function(i, self.input_data, self.vr_data, self.function_params)

    # redefine post init
    def __postinit__(self) -> None:
        """Do some housekeeping for the process class"""

        self.__misc_init__()

    def get_process_args(self) -> tuple:
        """return the data associated with this object"""

        self.func_name: function = self.function

        return (
            self.func_name,
            self.input_data,
            self.vr_data,
            self.function_params,
        )


class AddSignal(Process):
    """This process adds values to the current flux based on the values provided by the signal object.
    This class is typically invoked through the connector object

     Example::

     AddSignal(name = "name",
               reservoir = upstream_reservoir_handle,
               flux = flux_to_act_upon,
               lt = flux with lookup values)

     where - the upstream reservoir is the reservoir the process belongs too
             the flux is the flux to act upon
             lt= contains the flux object we lookup from

    """

    def __init__(self, **kwargs: Dict[str, any]) -> None:
        """
        Create a new process object with a given process type and options
        """

        # get default names and update list for this Process
        self.__defaultnames__()  # default kwargs names
        self.lrk.extend(["lt", "flux", "reservoir"])  # new required keywords

        self.__initerrormessages__()
        # self.bem.update({"rate": "a string"})
        self.__validateandregister__(kwargs)  # initialize keyword values

        # legacy variables
        self.mo = self.reservoir.mo
        self.__postinit__()  # do some housekeeping
        self.__register_name__()

        # defaults
        self.__execute__ = self.__add_with_fi__
        self.__get_process_args__ = self.__get_process_args_fi__

    # setup a placeholder call function
    def __call__(self, i: int):
        return self.__execute__(i)

    # use this when we do isotopes
    def __add_with_fi__(self, i) -> None:
        """Each process is associated with a flux (self.f). Here we replace
        the flux value with the value from the signal object which
        we use as a lookup-table (self.lt)

        Note that the signal may also specify a delta value. Thus we
        need to
        1) get the mass of flux + signal
        2) add the delta of flux and signal
        3) calculate the new li and hi

        """

        # add signal mass to flux mass
        r = self.f.species.r
        fm = self.f.m[i]  # flux rate
        fl = self.f.l[i]  # flux rate light isotope
        sm = self.lt.m[i]  # signal mass
        sl = self.lt.l[i]  # signal li

        # get signal delta
        if sm > 0:
            sd = 1000 * ((sm - sl) / sl - r) / r
        else:
            sd = 0
        # print(f"Signal delta = {sd}")

        if fm > 0:  # get flux delta
            fd = 1000 * ((fm - fl) / fl - r) / r + sd
        else:
            fd = sd

        fm += sm  # add signal mass
        fl = 1000.0 * fm / ((fd + 1000.0) * r + 1000.0)

        self.f.fa = np.array([fm, fl])
        # print(f"fa = {self.f.fa}\n")

    def __add_with_fa__(self, i) -> None:
        """same as above but use fa instead of flux data"""

        # add signal mass to flux mass
        r = self.f.species.r
        fm = self.f.fa[0]  # flux rate
        fl = self.f.fa[1]  # flux rate light isotope
        sm = self.lt.m[i]  # signal mass
        sl = self.lt.l[i]  # signal li

        # get signal delta
        sd = 1000 * ((sm - sl) / sl - r) / r

        if fm > 0:  # get flux delta
            fd = 1000 * ((fm - fl) / fl - r) / r + sd
        else:
            fd = sd

        # set new flux rate
        fm += sm  # add signal mass
        fl = 1000.0 * fm / ((fd + 1000.0) * r + 1000.0)

        self.f.fa = np.array([fm, fl])

    def get_process_args(self):
        return self.__get_process_args__()

    def __get_process_args_fi__(self):

        func_name: function = self.p_add_signal_fi

        print(f"flux_name = {self.flux.full_name}")

        data = List(
            [
                self.flux.m,  # 0
                self.flux.l,  # 1
                self.lt.m,  # 2
                self.lt.l,  # 3
                self.flux.fa,  # 4
            ]
        )

        params = List([float(self.reservoir.species.element.r)])

        return func_name, data, params

    @staticmethod
    @njit(fastmath=True, error_model="numpy")
    def p_add_signal_fi(data, params, i) -> None:

        r: float = params[0]
        fm: float = data[0][i]  # fm
        fl: float = data[1][i]  # fl
        sm: float = data[2][i]  # sm
        sl: float = data[3][i]  # sd

        # get signal delta
        if sm > 0:
            sd = 1000 * ((sm - sl) / sl - r) / r
        else:
            sd = 0

        if fm > 0:  # get flux delta
            fd = 1000 * ((fm - fl) / fl - r) / r + sd
        else:
            fd = sd

        # set new flux rate
        fm += sm  # add signal mass
        fl = 1000.0 * fm / ((fd + 1000.0) * r + 1000.0)

        data[4][:] = [fm, fl]

    def __get_process_args_fa__(self):

        func_name: function = self.p_add_signal_fa

        print(f"flux_name = {self.flux.full_name}")

        data = List(
            [
                self.lt.m,  # 0
                self.lt.l,  # 1
                self.flux.fa,  #  2
            ]
        )

        params = List([float(self.reservoir.species.element.r)])

        return func_name, data, params

    @staticmethod
    @njit(fastmath=True, error_model="numpy")
    def p_add_signal_fa(data, params, i) -> None:

        r: float = params[0]
        fm: float = data[2][0]
        fl: float = data[2][1]
        sm: float = data[0][i]
        sl: float = data[1][i]

        # get signal delta
        if sm > 0:
            sd = 1000 * ((sm - sl) / sl - r) / r
        else:
            sd = 0

        if fm > 0:  # get flux delta
            fd = 1000 * ((fm - fl) / fl - r) / r + sd
        else:
            fd = sd

        fm += sm  # add signal mass
        fl = 1000.0 * fm / ((fd + 1000.0) * r + 1000.0)
        data[2][:] = [fm, fl]


class SaveFluxData(Process):
    """
    This process stores the flux data from each iteration into a vector
    Example::

         SaveFluxData(name = "Name",
                   flux = Flux Handle
    )

    """

    __slots__ = "flux"

    def __init__(self, **kwargs: Dict[str, any]) -> None:
        """Initialize this Process"""
        # get default names and update list for this Process
        self.__defaultnames__()  # default kwargs names
        self.lrk.extend(["flux"])  # new required keywords

        self.__validateandregister__(kwargs)  # initialize keyword values

        # legacy variables
        self.__postinit__()  # do some housekeeping
        self.__register_name__()

    # setup a placeholder call function
    def __call__(self, i: int):

        self.f[i] = self.f.fa

    def get_process_args(self):
        """"""

        func_name: function = self.p_save_flux

        data = List(
            [
                self.flux.m,  # 0
                self.flux.l,  # 1
                # self.flux.h,  # 2
                # self.flux.d,  # 3
                self.flux.fa,  # 4 2
            ]
        )

        params = List([0.0, 0.0, 0.0])

        return func_name, data, params

    @staticmethod
    @njit(fastmath=True, error_model="numpy")
    def p_save_flux(data, params, i) -> None:

        data[0][i] = data[2][0]
        data[1][i] = data[2][1]


class ScaleFlux(Process):
    """This process scales the mass of a flux (m,l,h) relative to
    another flux. Delta is either taken from the upstream reserevoir
    or set to a fixed value.  The scale factor "scale" and flux
    reference must be present when the object is being initalized

    Example::
         ScaleFlux(name = "Name",
                   reservoir = reservoir_handle (upstream or downstream)
                   scale = 1
                   ref_flux = flux we use for scale
                   )

    """

    __slots__ = ("rate", "scale", "ref_reservoirs", "reservoir", "flux")

    def __init__(self, **kwargs: Dict[str, any]) -> None:
        """Initialize this Process"""
        # get default names and update list for this Process
        self.__defaultnames__()  # default kwargs names
        self.lrk.extend(["reservoir", "flux", "scale"])  # new required keywords

        self.__validateandregister__(kwargs)  # initialize keyword values

        if "ref_reservoirs" in kwargs:
            self.ref_flux = kwargs["ref_reservoirs"]
        elif "ref_flux" in kwargs:
            pass
        else:
            raise ValueError("You need to specify a value for ref_flux")

        # legacy variables
        self.mo = self.reservoir.mo
        self.__postinit__()  # do some housekeeping
        self.__register_name__()

        """ Decide how to calculate isotopes (if)
        If the connection specifies delta explicitly, this will
        override any settings in the source.
        """
        self.c = 1
        if self.delta != "None":
            # delta explicitly provided
            l = get_l_mass(1, self.delta, self.reservoir.species.element.r)
            self.c = l / (1 - l)
            self.__execute__ = self.__with_fixed_delta__
        elif isinstance(self.source, Source):
            if self.source.delta != "None":
                self.delta = self.source.delta
                l = get_l_mass(1, self.delta, self.reservoir.species.element.r)
                self.c = l / (1 - l)
                self.__execute__ = self.__with_fixed_delta__
            else:
                self.__execute__ = self.__without_isotopes__
        else:
            if self.source.isotopes:
                self.__execute__ = self.__with_isotopes__
            elif not self.source.isotopes:
                self.__execute__ = self.__without_isotopes__
            else:
                raise ValueError("This should never happen!")

    # create stubs
    def __with_source__():
        raise NotImplementedError()

    def __with_isotopes__():
        raise NotImplementedError()

    def __without_isotopes__():
        raise NotImplementedError()

    # setup a placeholder call function
    def __call__(self, i: int):
        return self.__execute__(i)

    def __with_isotopes__(self, i: int) -> None:
        """Apply the scale factor. This is typically done through the the
        model execute method.
        Note that this will use the mass of the flux we use for scaling, but that we will set the
        delta according to reservoir this flux derives from

        """

        # get reference flux
        rm: float = self.ref_flux.fa[0] * self.scale
        r: float = self.reservoir.species.element.r

        # get the target isotope ratio based on upstream delta
        c = self.reservoir.l[i - 1] / (
            self.reservoir.m[i - 1] - self.reservoir.l[i - 1]
        )

        fl: float = rm * c / (c + 1)

        self.flux.fa[:] = [rm, fl]

    def __with_fixed_delta__(self, i: int) -> None:
        """similar to with isotopes, but this time we take the
        isotope value of the source (which has no array, just a fixed
        value"""

        # get reference flux
        rm: float = self.ref_flux.fa[0] * self.scale
        r: float = self.reservoir.species.element.r

        # get the target isotope ratio based on upstream delta
        c = self.c

        fl: float = rm * c / (c + 1)

        self.flux.fa[:] = [rm, fl]

    def __without_isotopes__(self, i: int) -> None:
        """Apply the scale factor. This is typically done through the the
        model execute method.
        Note that this will use the mass of the reference object, but that we will set the
        delta according to the reservoir (or the flux?)

        """

        self.f.fa = np.array(
            [
                self.ref_flux.fa[0] * self.scale,
                0,
            ]
        )

    def get_process_args(self):
        """we need to decide between cases where the reservoir is an
        actual reserevoir or just a source, since sources do not have
        arrays

        """

        data = List(
            [
                self.flux.m,  # 0
                self.flux.l,  # 1
                self.ref_flux.m,  # 4 2 Reference Flux
                self.reservoir.m,  # 5 3 Upstream reservoir
                self.reservoir.l,  # 6 4 Upstream reservoir li
                self.flux.fa,  # 8 5
                self.ref_flux.fa,  # 9 6
            ]
        )

        if isinstance(self.source, Source) and self.source.isotopes:
            func_name: function = self.p_scale_flux_fd
        elif self.delta != "None":
            func_name: function = self.p_scale_flux_fd
        else:
            func_name: function = self.p_scale_flux_r

        params = List(
            [
                float(self.reservoir.species.element.r),
                float(self.scale),
                float(self.c),
            ]
        )

        return func_name, data, params

    @staticmethod
    @njit(fastmath=True, error_model="numpy")
    def p_scale_flux_r(data, params, i) -> None:
        """delta is derived from upstream reservoir"""

        # params
        r: float = params[0]  # r value
        s: float = params[1]  # scale

        # data
        mf: float = data[6][0] * s  # mass of reference flux
        mr: float = data[3][i - 1]  # mass upstream reserevoir
        lr: float = data[4][i - 1]  # li upstream reserevoir

        # get the target isotope ratio based on upstream delta
        c = lr / (mr - lr)
        l = mf * c / (c + 1)

        data[5][:] = [mf, l]

    @staticmethod
    @njit(fastmath=True, error_model="numpy")
    def p_scale_flux_fd(data, params, i) -> None:
        """delta is set to a fixed value"""

        # params
        r: float = params[0]  # r value
        s: float = params[1]  # scale
        c: float = params[2]  # l/h ratio

        # data
        mf: float = data[6][0] * s  # mass of reference flux
        mr: float = data[3][i - 1]  # mass upstream reserevoir
        lr: float = data[4][i - 1]  # li upstream reserevoir

        # get the target isotope ratio based on upstream delta
        l = mf * c / (c + 1)

        data[5][:] = [mf, l]


class Fractionation(Process):
    """This process offsets the isotopic ratio of the flux by a given
       delta value. In other words, we add a fractionation factor

    Example::
         Fractionation(name = "Name",
                       reservoir = upstream_reservoir_handle,
                       flux = flux handle
                       alpha = 12 in permil (e.f)

    """

    __slots__ = ("flux", "reservoir")

    def __init__(self, **kwargs: Dict[str, any]) -> None:
        """Initialize this Process"""
        # get default names and update list for this Process
        self.__defaultnames__()  # default kwargs names
        self.lrk.extend(["reservoir", "flux", "alpha"])  # new required keywords

        self.__validateandregister__(kwargs)  # initialize keyword values
        self.__postinit__()  # do some housekeeping

        # alpha is given in permil, but the fractionation routine expects
        # it as 1 + permil, i.e., 70 permil would 1.007

        self.alp = 1 + self.alpha / 1000
        self.mo = self.reservoir.mo
        self.__register_name__()

    def __call__(self, i: int) -> None:
        """
        Set flux isotope masses based on fractionation factor
        relative to reserevoir

        """

        # print(f"self.f.m[i] =  {self.f.m[i]}")
        r = self.reservoir.species.element.r
        fm = self.f.fa[0]
        rm = self.reservoir.m[i - 1]
        rl = self.reservoir.l[i - 1]
        a = self.alp

        if fm != 0:
            fl = rl * fm / (a * rm + rl - a * rl)
            self.f.fa[:] = [fm, fl]
        else:
            self.f.fa[:] = [
                0,
                0,
            ]

        return

    def get_process_args(self):

        func_name: function = self.p_fractionation

        data = List(
            [
                self.flux.m,  # 0
                self.flux.l,  # 1
                self.reservoir.m,  # 4 2
                self.reservoir.l,  # 5 3
                self.flux.fa,  # 6 4
            ]
        )
        params = List([float(self.reservoir.species.element.r), float(self.alp)])

        return func_name, data, params

    @staticmethod
    @njit(fastmath=True, error_model="numpy")
    def p_fractionation(data, params, i) -> None:
        # params
        r: float = params[0]  # rvalue
        a: float = params[1]  # alpha

        # data
        fm: float = data[4][0]  # flux mass
        rm: float = data[2][i - 1]  # 4 reservoir mass
        rl: float = data[3][i - 1]  # 4 reservoir light isotope

        fl = rl * fm / (a * rm + rl - a * rl)
        data[4][:] = [fm, fl]


class RateConstant(Process):
    """This is a wrapper for a variety of processes which depend on rate constants
    Please see the below class definitions for details on how to call them
    At present, the following processes are defined



    """

    __slots__ = ("scale", "ref_value", "k_value", "flux", "reservoir")

    def __init__(self, **kwargs: Dict[str, any]) -> None:
        """Initialize this Process"""

        from . import ureg, Q_
        from .connections import SourceGroup, SinkGroup, ReservoirGroup
        from .connections import ConnectionGroup, GasReservoir
        from esbmtk import Flux

        # Note that self.lkk values also need to be added to the lkk
        # list of the connector object.

        # get default names and update list for this Process
        self.__defaultnames__()  # default kwargs names

        # update the allowed keywords
        self.lkk: dict = {
            "scale": (Number, np.float64),
            "k_value": (Number, np.float64),
            "name": str,
            "reservoir": (Reservoir, Source, Sink, GasReservoir, np.ndarray),
            "source": (Reservoir, Source, GasReservoir),
            "flux": Flux,
            "ref_reservoirs": list,
            "reservoir_ref": (Reservoir, GasReservoir),
            "left": (list, Reservoir, Number, np.float64, np.ndarray),
            "right": (list, Reservoir, Number, np.ndarray),
            "register": (
                SourceGroup,
                SinkGroup,
                ReservoirGroup,
                ConnectionGroup,
                Flux,
                str,
            ),
            "gas": (Reservoir, GasReservoir, Source, Sink, np.ndarray),
            "liquid": (Reservoir, Source, Sink),
            "solubility": (Number, np.float64),
            "piston_velocity": (Number, np.float64),
            "water_vapor_pressure": (Number, np.float64),
            "ref_species": np.ndarray,
            "seawaterconstants": any,
            "isotopes": bool,
            "function_reference": any,
            "f_0": (str),
            "pco2_0": (str),
            "ex": Number,
            "delta": (Number, str),
        }

        # new required keywords
        self.lrk.extend([["reservoir", "atmosphere"], ["scale", "k_value"]])

        # dict with default values if none provided
        # self.lod = {r
        self.lod: dict = {
            "isotopes": False,
            "function_reference": "None",
            "delta": "None",
        }

        self.__initerrormessages__()

        # add these terms to the known error messages
        self.bem.update(
            {
                "scale": "a number",
                "reservoir": "Reservoir handle",
                "ref_reservoirs": "List of Reservoir handle(s)",
                "ref_value": "a number or flux quantity",
                "name": "a string value",
                "flux": "a flux handle",
                "left": "list, reservoir or number",
                "right": "list, reservoir or number",
                "function_reference": "A function reference",
            }
        )

        # initialize keyword values
        self.__validateandregister__(kwargs)
        if "reservoir" in kwargs:
            self.mo = self.reservoir.mo
        elif "gas_reservoir" in kwargs:
            self.mo = self.gas_reservoir

        self.__misc_init__()
        self.__postinit__()  # do some housekeeping
        # legacy variables

        self.__register_name__()

        if self.reservoir.isotopes or self.isotopes:
            self.__execute__ = self.__with_isotopes__
            self.c = 1
            if self.delta != "None":
                self.__execute__ = self.__with_fixed_delta__
                l = get_l_mass(1, self.delta, self.reservoir.species.r)
                self.c = l
                # print(f"delta = {self.delta} setting c = {self.c}")
        else:
            self.__execute__ = self.__without_isotopes__
            self.c = 1

    def __postinit__(self) -> "None":
        self.mo = self.reservoir.mo

    # setup a placeholder call function
    def __call__(self, i: int):
        return self.__execute__(i)


class weathering(RateConstant):
    """This process calculates the flux as a function of the upstream
     reservoir concentration C and a constant which describes thet
     strength of relation between the reservoir concentration and
     the flux scaling

     F = f_0 * (scale * C/pco2_0)**ncc

     where C denotes the concentration in the ustream reservoir, k is a
     constant. This process is typically called by the connector
     instance. However you can instantiate it manually as

     weathering(
                       name = "Name",
                       reservoir = upstream_reservoir_handle,
                       reservoir_ref = reference_reservoir (Atmosphere)
                       flux = flux handle,
                       ex = exponent
                       pco2_0 = 280,
                       f_0 = 12 / 17e12
                       Scale =  1000,
                       delta = 0,

    )

    """

    def __misc_init__(self):
        """
        Scale the flux relative to the pco2_0 concentration.
        The delta values are derived from either:

        - the upstream reservoir
        - the upstream source
        - set to a fixed value

        """

        self.f_0: float = Q_(self.f_0).to(self.mo.f_unit).magnitude
        self.pco2_0 = Q_(self.pco2_0).to("ppm").magnitude * 1e-6

        # if self.delta == "None" and  self.source.isotopes == False:
        #      self.fixed = True
        # delta provided explicitly
        if self.delta != "None":
            self.d = self.delta
            self.isotopes = True
            l = get_l_mass(1, self.d, self.source.species.element.r)
            self.c = l / (1 - l)
            self.func_name = self.p_weathering_fd
            self.__execute__ = self.__with_fixed_delta__
            self.fixed = True

        # source is Source and has delta
        elif self.source.isotopes and isinstance(self.source, Source):
            self.isotopes = True
            self.delta = self.source.delta
            self.func_name = self.p_weathering_fd
            self.__execute__ = self.__with_fixed_delta__
            self.c = self.source.c
            self.fixed = True

        # source is reservoir
        elif self.source.isotopes:
            self.isotopes = True
            self.func_name = self.p_weathering
            self.__execute__ = self.__with_isotopes__
            self.fixed = False

        # source is source, has no delta, and delta is not provided
        else:
            self.c = 1
            self.isotopes = False
            self.func_name = self.p_weathering_fd
            self.__execute__ = self.__without_isotopes__
            self.fixed = True

    def __without_isotopes__(self, i: int) -> None:

        f = self.scale * (
            self.f_0 * (self.reservoir_ref.c[i - 1] / self.pco2_0) ** self.ex
        )
        self.flux.fa = np.array([f, 0])

    def __with_isotopes__(self, i: int) -> None:
        """
        C = M/V so we express this as relative to mass which allows us to
        use the isotope data.

        The below calculates the flux as function of reservoir concentration,
        rather than scaling the flux.
        """

        c = self.flux.fa[1] / self.flux.fa[0]
        f = self.scale * (
            self.f_0 * (self.reservoir_ref.c[i - 1] / self.pco2_0) ** self.ex
        )
        fl = f * c
        self.flux.fa = np.array([f, fl])

    def __with_fixed_delta__(self, i: int) -> None:
        """ """

        f = self.scale * (
            self.f_0 * (self.reservoir_ref.c[i - 1] / self.pco2_0) ** self.ex
        )
        fl = f * self.c
        self.flux.fa = np.array([f, fl])

    def get_process_args(self):
        """data depends on whether upstream, is source or
        reserevoir/gas-reservoir"""

        if self.fixed:
            data = List(
                [
                    self.flux.fa,  # 0
                    self.reservoir_ref.c,  # 1
                ]
            )
        else:
            data = List(
                [
                    self.flux.fa,  # 0
                    self.reservoir_ref.c,  # 1
                    self.source.m,  # 2
                    self.source.l,  # 3
                ]
            )

        params = List(
            [
                float(self.reservoir.species.element.r),  # 0
                float(self.scale),  # 1
                float(self.f_0),  # 2
                float(self.pco2_0),  # 3
                float(self.ex),  # 4
                float(self.c),  # 5
            ]
        )

        return self.func_name, data, params

    @staticmethod
    @njit(fastmath=True, error_model="numpy")
    def p_weathering(data, params, i) -> None:
        """ delta value depends on upstream reservoir """ 
        # params
        s: float = params[1]
        f_0: float = params[2]
        pco2_0: float = params[3]
        ex: float = params[4]

        # data
        cr = data[3][i - 1] / data[2][i - 1]
        c: float = data[1][i - 1]
        
        f: float = s * f_0 * (c / pco2_0) ** ex
        fl: float = f * cr
        data[0][:] = [f, fl]

    @staticmethod
    @njit(fastmath=True, error_model="numpy")
    def p_weathering_fd(data, params, i) -> None:
        """ delta value is fixed """ 
        # params
        s: float = params[1]
        f_0: float = params[2]
        pco2_0: float = params[3]
        ex: float = params[4]
        cr = params[5]

        c: float = data[1][i - 1]
        f: float = s * f_0 * (c / pco2_0) ** ex
        fl = f * cr
        data[0][:] = [f, fl]


class MultiplySignal(Process):
    """This process mulitplies a given flux with the the data in the
    signal.  This class is typically invoked through the connector
    object: Note that this process will not modify the delta value of
    a given flux.  If you needto vary the delta value it is best to
    add a second signal which uses the add signal type.

     Example::
     MultiplySignal(name = "name",
               reservoir = upstream_reservoir_handle,
               flux = flux_to_act_upon,
               lt = flux with lookup values)

     where the upstream reservoir is the reservoir the process belongs
             too the flux is the flux to act upon lt= contains the
             flux object we lookup from

    """

    def __init__(self, **kwargs: Dict[str, any]) -> None:
        """
        Create a new process object with a given process type and options
        """

        # get default names and update list for this Process
        self.__defaultnames__()  # default kwargs names
        self.lrk.extend(["lt", "flux", "reservoir"])  # new required keywords

        self.__initerrormessages__()
        # self.bem.update({"rate": "a string"})
        self.__validateandregister__(kwargs)  # initialize keyword values

        # legacy variables
        self.mo = self.reservoir.mo
        self.__postinit__()  # do some housekeeping
        self.__register_name__()

        # decide whichh call function to use
        # if self.mo.m_type == "both":

        # default
        self.__execute__ = self.__multiply_with_flux_fi__
        self.__get_process_args__ = self.__get_process_args_fi__

    # setup a placeholder call function
    def __call__(self, i: int):
        return self.__execute__(i)

    # use this when we do isotopes
    def __multiply_with_flux_fi__(self, i) -> None:
        """Each process is associated with a flux (self.f). Here we replace
        the flux value with the value from the signal object which
        we use as a lookup-table (self.lt)
        """

        # multiply flux mass with signal
        c = self.lt.m[i]
        m = self.f.m[i] * c
        l = self.f.l[i] * c
        # h = self.f.h[i] * c
        # d = self.f.d[i]
        self.flux.fa = np.array([m, l])
        print(f"multiply fa {self.flux.fa}, f. = {self.f.m[i]}, c = {c} ")

    def __multiply_with_flux_fa__(self, i) -> None:
        """Each process is associated with a flux (self.f). Here we replace
        the flux value with the value from the signal object which
        we use as a lookup-table (self.lt)
        """

        # multiply flux mass with signal
        c = self.lt.m[i]
        m = self.f.fa[0] * c
        l = self.f.fa[1] * c
        self.flux.fa = np.array([m, l])

    def __get_process_args_fi__(self):
        func_name: function = self.p_multiply_signal_fi
        data = List(
            [
                self.flux.m,  # 0
                self.flux.l,  # 1
                # self.lt.l,  # 2
                # self.flux.d,  # 3
                self.lt.m,  # 2
                self.flux.fa,  # 3
            ]
        )
        params = List([float(self.reservoir.species.element.r)])

        return func_name, data, params

    def __get_process_args_fa__(self):
        func_name: function = self.p_multiply_signal_fa
        data = List(
            [
                self.lt.m,  # 0
                self.flux.fa,  # 1
            ]
        )
        params = List([float(self.reservoir.species.element.r)])

        return func_name, data, params

    def get_process_args(self):

        return self.__get_process_args__()

    @staticmethod
    @njit(fastmath=True, error_model="numpy")
    def p_multiply_signal_fi(data, params, i) -> None:

        r = params[0]

        c = data[2][i]
        m = data[0][i] * c
        l = data[1][i] * c

        data[3][:] = [m, l]

    @staticmethod
    @njit(fastmath=True, error_model="numpy")
    def p_multiply_signal_fa(data, params, i) -> None:

        r = params[0]

        c = data[0][i]
        m = data[1][0] * c  # m
        l = data[1][1] * c  # l

        data[1][:] = [m, l]


class VarDeltaOut(Process):
    """Unlike a passive flux, this process sets the flux istope ratio
    equal to the isotopic ratio of the reservoir. The
    init and register methods are inherited from the process
    class.
    VarDeltaOut(name = "name",
                reservoir = upstream_reservoir_handle,
                flux = flux handle,
                rate = rate,)
    """

    __slots__ = ("rate", "flux", "reservoir")

    def __init__(self, **kwargs: Dict[str, any]) -> None:
        """Initialize this Process"""

        from . import ureg, Q_
        from .connections import ConnectionGroup
        from esbmtk import Flux, Reservoir, ReservoirGroup

        # get default names and update list for this Process
        self.__defaultnames__()
        self.lkk: Dict[str, any] = {
            "name": str,
            "reservoir": (Reservoir, Source, Sink),
            "flux": Flux,
            "rate": (str, Q_),
            "register": (ConnectionGroup, ReservoirGroup, Reservoir, Flux, str),
            "scale": (Number, np.float64, str),
        }
        self.lrk.extend(["reservoir", "flux"])  # new required keywords
        self.__initerrormessages__()
        self.__validateandregister__(kwargs)  # initialize keyword values
        self.mo = self.reservoir.mo
        self.__postinit__()  # do some housekeeping
        self.__register_name__()

        # decide which call function to use
        # if self.mo.m_type == "both":
        if self.reservoir.isotopes:
            # print(
            #    f"vardeltaout with isotopes for {self.reservoir.register.name}.{self.reservoir.name}"
            # )
            if isinstance(self.reservoir, Reservoir):
                # print("Using reservoir")
                self.__execute__ = self.__with_isotopes_reservoir__
            elif isinstance(self.reservoir, Source):
                # print("Using Source")
                self.__execute__ = self.__with_isotopes_source__
            else:
                raise ValueError(
                    f"{self.name}, reservoir must be of type Source or Reservoir, not {type(self.reservoir)}"
                )
        else:
            self.__execute__ = self.__without_isotopes__

    # setup a placeholder call function
    def __call__(self, i: int):
        return self.__execute__(i)

    def __with_isotopes_reservoir__(self, i: int) -> None:
        """Here we re-balance the flux. This code will be called by the
        apply_flux_modifier method of a reservoir which itself is
        called by the model execute method
        """

        m: float = self.flux.m[i]
        if m != 0:
            c = self.reservoir.l[i - 1] / (
                self.reservoir.m[i - 1] - self.reservoir.l[i - 1]
            )
            self.flux.fa = [m, m * c]

    def get_process_args(self):
        """Provide the data structure which needs to be passed to the numba solver"""

        # if upstream is a source, we only have a single delta value
        # so we need to patch this. Maybe this should move to source?

        func_name: function = self.p_vardeltaout

        data = List(
            [
                self.flux.m,  # 0
                # self.flux.l,  # 1
                # self.flux.h,  # 2
                # self.flux.d,  # 3
                # delta,  # 4
                self.reservoir.m,  # 1
                self.reservoir.l,  # 2
                self.flux.fa,  # 5 3
            ]
        )

        params = List([float(reservoir.species.element.r)])

        return func_name, data, params

    @staticmethod
    @njit(fastmath=True, error_model="numpy")
    def p_vardeltaout(data, params, i) -> None:
        # concentration times scale factor

        mf: float = data[0][i - 1]  # flux mass
        mr: float = data[1][i - 1]  # reservoir mass
        lr: float = data[2][i - 1]  # reservoir l

        c = lr / (mr - lr)
        data[3][:] = [mf, mf * c]

    def __with_isotopes_source__(self, i: int) -> None:
        """If the source of the flux is a source, there is only a single delta value.
        Changes to the flux delta are applied through the Signal class.
        """

        m: float = self.flux.m[i]
        if m != 0:
            c = self.reservoir.l[i - 1] / (
                self.reservoir.m[i - 1] - self.reservoir.l[i - 1]
            )
            self.flux.fa = [m, m * c]

    def __without_isotopes__(self, i: int) -> None:
        """Here we re-balance the flux. This code will be called by the
        apply_flux_modifier method of a reservoir which itself is
        called by the model execute method
        """

        raise NotImplementedError("vardeltaout w/o isotopes is not defined")


class ScaleRelativeToMass(RateConstant):
    """This process scales the flux as a function of the upstream
     reservoir Mass M and a constant which describes the
     strength of relation between the reservoir mass and
     the flux scaling

     F = F0 *  M * k

     where M denotes the mass in the ustream reservoir, k is a
     constant and F0 is the initial unscaled flux. This process is
     typically called by the connector instance. However you can
     instantiate it manually as

     Note that we scale the flux, rather than compute the flux!

     This is faster than setting a new flux, computing the isotope
     ratio and setting delta. So you either have to set the initial
     flux F0 to 1, or calculate the scale accordingly

     ScaleRelativeToMass(
                       name = "Name",
                       reservoir= upstream_reservoir_handle,
                       flux = flux handle,
                       Scale =  1000,
    )

    """

    def __without_isotopes__(self, i: int) -> None:
        m: float = self.reservoir.m[i - 1] * self.scale
        self.flux.fa = [m, 0, 0, 0]

    def __with_isotopes__(self, i: int) -> None:
        """
        this will be called by the Model.run() method

        """
        m: float = self.reservoir.m[i - 1] * self.scale
        r: float = self.reservoir.species.element.r
        # d: float = self.reservoir.d[i - 1]
        c = self.reservoir.l[i - 1] / (
            self.reservoir.m[i - 1] - self.reservoir.l[i - 1]
        )
        l = m * c / (c + 1)
        self.flux.fa[:] = [m, l]

    def get_process_args(self):
        """return the data associated with this object"""

        func_name: function = self.p_scale_relative_to_mass

        data = List(
            [
                self.flux.m,  # 0
                self.flux.l,  # 1
                self.reservoir.m,  # 4 2
                self.reservoir.l,  # 5 3
                self.flux.fa,  # 7 4
            ]
        )

        params = List([float(self.reservoir.species.element.r), float(self.scale)])

        return func_name, data, params

    @staticmethod
    @njit(fastmath=True, error_model="numpy")
    def p_scale_relative_to_mass(data, params, i) -> None:
        # concentration times scale factor
        # params
        r: float = params[0]  # r values
        s: float = params[1]  # scale factor

        # data
        rm = data[2][i - 1]
        rl = data[3][i - 1]
        # rd = data[6][i - 1]

        m: float = rm * s  # new flux
        c: float = rl / (rm - rl)
        l: float = m * c / (c + 1)
        # d: float = data[1][i - 1]  # flux delta

        data[4][:] = [m, l]


class GasExchange(RateConstant):
    """

    GasExchange(
          gas =GasReservoir, #
          liquid = Reservoir, #,
          ref_species = array of concentrations #
          solubility=Atmosphere.swc.SA_co2 [mol/(m^3 atm)],
          area = area, # m^2
          piston_velocity = m/year
          seawaterconstants = Ocean.swc
          water_vapor_pressure=Ocean.swc.p_H2O,
    )


    """

    # redefine misc_init which is being called by post-init
    def __misc_init__(self):
        """Set up input variables"""

        self.p_H2O = self.seawaterconstants.p_H2O
        self.a_dg = self.seawaterconstants.a_dg
        self.a_db = self.seawaterconstants.a_db
        self.a_u = self.seawaterconstants.a_u
        self.rvalue = self.liquid.sp.r
        self.volume = self.gas.volume

    def __without_isotopes__(self, i: int) -> None:

        # set flux
        # note that the sink delta is co2aq as returned by the carbonate VR
        # this equation is for mmol but esbmtk uses mol, so we need to
        # multiply by 1E3

        f = self.scale * (  # area in m^2
            self.gas.c[i - 1]  #  Atmosphere
            * (1 - self.p_H2O)  # p_H2O
            * self.solubility  # SA_co2 = mol/(m^3 atm)
            - self.ref_species[i - 1] * 1000  # [CO2]aq mol
        )
        # print(self.gas.c[i - 1])

        # changes in the mass of CO2 also affect changes in the total mass
        # of the atmosphere. So we need to update the reservoir volume
        # variable which we use the store the total atmospheric mass
        """The solver will use f and f12 to update mass, light
        isotope and concentration values in the Gas
        reserevoir. However, Gas reserevoirs track total gas pressure
        in the volume variable. This will not be updated by the
        solver. So we need to do it ourseleves

        """
        # self.gas.v[i] = self.gas.v[i - 1] + f * self.gas.mo.dt
        self.flux.fa[:] = [a, 1]

    def __with_isotopes__(self, i: int) -> None:
        """
        In the following I assume near neutral pH between 7 and 9, so that
        the isotopic composition of HCO3- is approximately equal to the isotopic
        ratio of DIC. The isotopic ratio of [CO2]aq can then be obtained from DIC via
        swc.e_db (swc.a_db)

        The fractionation factor subscripts denote the following:

        g = gaseous
        d = dissolved
        b = bicarbonate ion
        c = carbonate ion

        a_db is thus the fractionation factor between dissolved CO2aq and HCO3-
        and a_gb between CO2g HCO3-

        ref_species = DIC.cs.CO2aq
        liquid.m = DIC.m (used to get the isotope ratio)

        gas.m = mass of CO2
        gas.l = mass of 12CO2
        gas.c = CO2 concentration
        gas.v = total mass of atmosphere

        """

        from esbmtk import get_delta

        r = self.liquid.sp.r

        # equilibrium concentration of CO2 in water based on pCO2
        eco2_at = (
            self.gas.c[i - 1]  # p Atmosphere
            * (1 - self.p_H2O)  # p_H2O
            * self.solubility
        )
        # equilibrium concentration of CO2 in water based on CO2aq
        eco2_aq = self.ref_species[i - 1] * 1000

        #  flux
        f = self.scale * (eco2_at - eco2_aq)

        c13g = 1 - self.gas.l[i - 1] / self.gas.m[i - 1]
        c13aq = 1 - self.liquid.l[i - 1] / self.liquid.m[i - 1]
        # get 13C CO2 equlibrium concentration  CO2 in water based on pCO2
        eco2_at_13 = (
            self.gas.c[i - 1]
            * c13g
            * (1 - self.p_H2O)  # p_H2O
            * self.solubility
            * self.a_dg
        )

        # get 13C equilibrium  CO2 in water based on DIC m & l
        eco2_aq_13 = self.a_db * eco2_aq * c13aq

        # l =  eco2_aq - eco2_aq_13
        # d = get_delta(l, eco2_aq_13, r)
        # print(f"d based on CO2aq = {d:.2f}")

        l = eco2_at - eco2_at_13
        d = get_delta(l, eco2_at_13, r)
        # print(f"d based on CO2at = {d:.2f}")

        # 13C flux
        f13 = self.scale * self.a_u * (eco2_at_13 - eco2_aq_13)
        f12 = f - f13

        # print(f"e13c at = {eco2_at_13:.2e} e13c aq = {eco2_aq_13:.2e}")
        # print(f"diff =  {eco2_at - eco2_aq:.2e}")
        # print(f"diff 13c =  {eco2_at_13 - eco2_aq_13:.2e}")
        # d = get_delta(f12, f13, r)
        # print(f"i = {i} d gex = {d:.2f}, f = {f:.2e}, f13: {f13:.2e} fr = {f/f12:.2e}\n")

        self.flux.fa = [f, f12]
        """The solver will use f and f12 to update mass, light
        isotope and concentration values in the Gas
        reserevoir. However, Gas reserevoirs track total gas pressure
        in the volume variable. This will not be updated by the
        solver. So we need to do it ourseleves

        not working, see note below
        """
        self.gas.v[i] = self.gas.v[i - 1] + f * self.gas.mo.dt
        # print()

    def __postinit__(self) -> None:
        """Do some housekeeping for the process class"""

        # legacy name aliases
        self.n: str = self.name  # display name of species
        self.r = self.liquid
        self.reservoir = self.liquid

    def get_process_args(self):
        """return the data associated with this object"""

        func_name: function = self.p_gas_exchange

        data = List(
            [
                self.flux.fa,  # 0
                self.liquid.m,  # 1
                self.liquid.l,  # 2
                self.gas.m,  # 3
                self.gas.l,  # 4
                self.gas.c,  # 5
                self.gas.v,  # 6
                self.ref_species,  # 7
            ]
        )

        params = List(
            [
                float(self.scale),  # 0
                float(self.solubility * (1 - self.p_H2O)),  # 1
                float(self.gas.volume),  # 3 2
                float(self.a_u),  # 4 3
                float(self.a_dg),  # 5 4
                float(self.a_db),  # 6 5
                float(self.gas.mo.dt),  # 7
            ]
        )

        self.params = params
        return func_name, data, params

    @staticmethod
    @njit(fastmath=True, error_model="numpy")
    def p_gas_exchange(data, params, i) -> None:
        """the below equation moved as many constants as possible
        outside of the function compared to the
        __with/without_isotopes__ method(s). See the
        __get_process_args__ method for details

        """

        scale: float = params[0]  # scale
        SA: float = params[1]  # solubility as  (1 - pH2O) * S
        gv: float = params[2]  # gas volume
        au: float = params[3]  #
        dg: float = params[4]  #
        db: float = params[5]  #
        dt: float = params[6]

        lm = data[1][i - 1]  # mass in liquid
        ll = data[2][i - 1]  # mass of light isotope in liquid_reservoir
        gm = data[3][i - 1]  # mass in gas reserevoir
        gl = data[4][i - 1]  # mass of light isotope in gas reserevoir
        gc = data[5][i - 1]  # concentration in gas reserevoir
        gv = data[6][i - 1]  # total mass of atmosphere
        rs = data[7][i - 1]  # concentrartion of reference species in liquid_reservoir
        gh = gm - gl  # mass of heavy isotope in gas reserevoir
        lh = lm - ll  # mass of heavy isotope in liquid reserevoir

        # equilibrium concentration of CO2 in water based on pCO2
        eco2_at = gc * SA  # p Atmosphere  # p_H2O
        # equilibrium concentration of CO2 in water based on CO2aq
        eco2_aq = rs * 1000
        f = scale * (eco2_at - eco2_aq)
        # get 13C CO2 equlibrium concentration
        eco2_at_13 = gh / gv * SA * dg  # p_H2O
        # get 13C in CO2aq
        eco2_aq_13 = db * eco2_aq * lh / lm
        # flux od 13C
        f13 = scale * au * (eco2_at_13 - eco2_aq_13)
        f12 = f - f13
        data[0][:] = [f, f12]  # fa
        """The solver will use f and f12 to update mass, light
        isotope and concentration values in the Gas
        reservoir. However, Gas reserevoirs track total gas pressure
        in the volume variable. This will not be updated by the
        solver. So we need to do it ourseleves

        This is currently not working, as the total mass increases faster
        than the species mass
        """
        # data[6][i] = data[6][i - 1] + f * dt * -1


class ScaleRelativeToConcentration(RateConstant):
    """This process calculates the flux as a function of the upstream
     reservoir concentration C and a constant which describes thet
     strength of relation between the reservoir concentration and
     the flux scaling

     F = C * k

     where C denotes the concentration in the ustream reservoir, k is a
     constant. This process is typically called by the connector
     instance. However you can instantiate it manually as

     ScaleRelativeToConcentration(
                       name = "Name",
                       reservoir= upstream_reservoir_handle,
                       flux = flux handle,
                       Scale =  1000,
                       delta = "None",
    )

    If delta is given the fluxes uses a ficed delta, If not it uses
    the upestream delta

    """

    def __without_isotopes__(self, i: int) -> None:
        m: float = self.reservoir.m[i - 1]
        if m > 0:  # otherwise there is no flux
            # convert to concentration
            c = m / self.reservoir.volume
            f = c * self.scale
            self.flux.fa = [
                f,
                0,
            ]

    def __with_isotopes__(self, i: int) -> None:
        """
        C = M/V so we express this as relative to mass which allows us to
        use the isotope data.

        The below calculates the flux as function of self.reservoir concentration,
        rather than scaling the flux.
        """

        rc: float = self.reservoir.c[i - 1]
        if rc > 0:  # otherwise there is no flux
            r: float = self.reservoir.species.element.r
            c = self.reservoir.l[i - 1] / (
                self.reservoir.m[i - 1] - self.reservoir.l[i - 1]
            )
            m = rc * self.scale
            l = m * c / (c + 1)
            self.f.fa = [m, l]

    def __with_fixed_delta__(self, i: int) -> None:
        """
        C = M/V so we express this as relative to mass which allows us to
        use the isotope data.

        The below calculates the flux as function of self.reservoir concentration,
        rather than scaling the flux.
        """

        rc: float = self.reservoir.c[i - 1]
        if rc > 0:  # otherwise there is no flux
            r: float = self.reservoir.species.element.r
            c = self.c
            m = rc * self.scale
            l = m * c / (c + 1)
            self.f.fa = [m, l]

    def get_process_args(self):

        if self.delta != "None":
            func_name: function = self.p_scale_relative_to_concentration_fd
            self.m = 1
            l = get_l_mass(1, self.delta, self.reservoir.species.element.r)
            self.c = l / (1 - l)
        else:
            func_name: function = self.p_scale_relative_to_concentration
            self.c = 1

        data = List(
            [
                self.flux.m,  # 0
                self.flux.l,  # 1
                self.reservoir.c,  # 5 2
                self.reservoir.m,  # 6 3
                self.reservoir.l,  # 7 4
                self.flux.fa,  # 8 5
            ]
        )
        params = List(
            [
                float(self.reservoir.species.element.r),
                float(self.scale),
                float(self.reservoir.volume),
                float(self.c),
            ]
        )

        return func_name, data, params

    @staticmethod
    @njit(fastmath=True, error_model="numpy")
    def p_scale_relative_to_concentration(data, params, i) -> None:
        """delta depends on upstream delta"""
        # params
        r: float = params[0]  # r
        s: float = params[1]  # scale factor
        v: float = params[2]  # res volume

        # # data 0 to 3 = flux data
        rc: float = data[2][i - 1]  # res concentration
        rm: float = data[3][i - 1]  # res mass
        rl: float = data[4][i - 1]  # res li

        if rc > 0:
            m: float = rc * s
            c: float = rl / (rm - rl)
            l: float = m * c / (c + 1)
            data[5][:] = [m, l]

        else:
            data[5][:] = [0.0, 0.0]

    @staticmethod
    @njit(fastmath=True, error_model="numpy")
    def p_scale_relative_to_concentration_fd(data, params, i) -> None:
        """delta depends fixed value"""
        # params
        r: float = params[0]  # r
        s: float = params[1]  # scale factor
        v: float = params[2]  # res volume
        c: float = params[3]  # isotope ratio

        # # data 0 to 3 = flux data
        rc: float = data[2][i - 1]  # res concentration
        rm: float = data[3][i - 1]  # res mass
        rl: float = data[4][i - 1]  # res li

        if rc > 0:
            m: float = rc * s
            l: float = m * c / (c + 1)
            data[5][:] = [m, l]

        else:
            data[5][:] = [0.0, 0.0]



"""
     esbmtk: A general purpose Earth Science box model toolkit
     Copyright (C), 2020 Ulrich G. Wortmann

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# from pint import UnitRegistry
from numbers import Number
from nptyping import NDArray
from typing import Dict, Union
from numpy import array, set_printoptions, arange, zeros, interp, mean
from pandas import DataFrame
from copy import deepcopy, copy
import time
from time import process_time
import numba
from numba.core import types
from numba import njit, prange
from numba.typed import List
from collections import OrderedDict

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

# import mpmath

import logging
import time

from esbmtk import Q_

# import builtins
# import math

set_printoptions(precision=4)


def insert_into_namespace(name, value, name_space=globals()):
    name_space[name] = value


def add_to(l, e):
    """
    add element e to list l, but check if the entry already exist. If so, throw
    exception. Otherwise add
    """

    if not (e in l):  # if not present, append element
        l.append(e)


def get_plot_layout(obj):
    """Simple function which selects a row, column layout based on the number of
    objects to display.  The expected argument is a reservoir object which
    contains the list of fluxes in the reservoir

    """

    noo = 1  # the reservoir is the fisrt object
    for f in obj.lof:  # count numbert of fluxes
        if f.plot == "yes":
            noo += 1

    for d in obj.ldf:  # count number of data fields
        noo += 1

    # noo = len(obj.lof) + 1  # number of objects in this reservoir
    logging.debug(f"{noo} subplots for {obj.n} required")

    size, geo = plot_geometry(noo)

    return size, geo


def plot_geometry(noo: int) -> tuple():
    """Define plot geometry based on number of objects to plot"""

    if noo < 2:
        geo = [1, 1]  # one row, one column
        size = [5, 3]  # size in inches
    elif 1 < noo < 3:
        geo = [2, 1]  # two rows, one column
        size = [5, 6]  # size in inches
    elif 2 < noo < 5:
        geo = [2, 2]  # two rows, two columns
        size = [10, 6]  # size in inches
    elif 4 < noo < 7:
        geo = [2, 3]  # two rows, three columns
        size = [15, 6]  # size in inches
    elif 6 < noo < 10:
        geo = [3, 3]  # two rows, three columns
        size = [15, 9]  # size in inches
    elif 9 < noo < 13:
        geo = [4, 3]  # two rows, three columns
        size = [15, 12]  # size in inches
    elif 12 < noo < 16:
        geo = [5, 3]  # two rows, three columns
        size = [15, 15]  # size in inches
    else:
        m = (
            "plot geometry for more than 15 fluxes is not yet defined"
            "Consider calling flux.plot individually on each flux in the reservoir"
        )
        raise ValueError(m)

    return size, geo


def list_fluxes(self, name, i) -> None:
    """
    Echo all fluxes in the reservoir to the screen
    """
    print(f"\nList of fluxes in {self.n}:")

    for f in self.lof:  # show the processes
        direction = self.lio[f.n]
        if direction == -1:
            t1 = "From:"
            t2 = "Outflux from"
        else:
            t1 = "To  :"
            t2 = "Influx to"

        print(f"\t {t2} {self.n} via {f.n}")

        for p in f.lop:
            p.describe()

    print(" ")
    for f in self.lof:
        f.describe(i)  # print out the flux data


def show_data(self, **kwargs) -> None:
    """Print the 3 lines of the data starting with index

    Optional arguments:

    index :int = 0 starting index
    indent :int = 0 indentation
    """

    off: str = "  "

    if "index" not in kwargs:
        index = 0
    else:
        index = kwargs["index"]

    if "indent" in kwargs:
        ind: str = kwargs["indent"] * " "
    else:
        ind: str = ""

    # show the first 4 entries
    # for i in range(index, index + 3):
    #     print(f"{off}{ind}i = {i}, Mass = {self.m[i]:.2e}, li= {self.l[i]:.2f}")


def set_y_limits(ax: plt.Axes, obj: any) -> None:
    """Prevent the display or arbitrarily small differences"""
    lower: float
    upper: float

    bottom, top = ax.get_ylim()
    if (top - bottom) < obj.display_precision:
        top = bottom + obj.display_precision
        ax.set_ylim(bottom, top)


def get_ptype(obj, **kwargs: dict) -> int:
    """
    Set plot type variable based on ptype or isotope keyword returns

    0 mass/concentration and isotope data
    1 isotopes only
    2 concentrartion only
    3 mass only

    """

    from esbmtk import Flux, Reservoir, Signal, DataField, Source, Sink

    ptype: int = 0

    if isinstance(obj, (Reservoir, Source, Sink, Flux)):
        if obj.isotopes:
            ptype = 0
        else:
            ptype = 2
    elif "ptype" in kwargs:
        if kwargs["ptype"] == "both":
            ptype = 0
        elif kwargs["ptype"] == "iso":
            ptype = 1
        elif kwargs["ptype"] == "concentration":
            ptype = 2
        elif kwargs["ptype"] == "mass_only":
            ptype = 2

    return ptype


def plot_object_data(geo: list, fn: int, obj: any) -> None:
    """collection of commands which will plotqand annotate a reservoir or flux
    object into an existing plot window.

    geo: geometry info
    fn: figure number in plot
    obj: the object to plot

    """

    from . import ureg, Q_
    from esbmtk import Flux, Reservoir, Signal, DataField, Source, GasReservoir

    # geo = list with rows and cols
    # fn  = figure number
    # yl  = array with y values for the left side
    # yr  = array with y values for the right side
    # obj = object handle, i.e., reservoir or flux

    first_axis: bool = False
    second_axis: bool = False

    rows = geo[0]
    cols = geo[1]
    # species = obj.sp
    model = obj.mo
    time = model.time + model.offset

    # convert data from model units to display units (i.e. the same
    # units the input data was defined).
    # time units are the same regardless of object
    time = (time * model.t_unit).to(model.d_unit).magnitude

    # we do not map isotope values
    if obj.isotopes:
        yr = obj.d

    # get plot type
    ptype: int = get_ptype(obj)

    # remap concentration & flux values
    if isinstance(obj, Flux):
        yl = (obj.m * model.f_unit).to(obj.plt_units).magnitude
        y_label = f"{obj.legend_left} [{obj.plt_units:~P}]"

    elif isinstance(obj, (Reservoir, GasReservoir)):
        if obj.display_as == "mass":
            yl = (obj.m * model.m_unit).to(obj.plt_units).magnitude
            y_label = f"{obj.legend_left} [{obj.plt_units:~P}]"
        elif obj.display_as == "ppm":
            yl = obj.c * 1e6
            y_label = f"ppm"
        elif obj.plot_transform_c != "None":
            if callable(obj.plot_transform_c):
                # yl = (obj.m * model.m_unit).to(obj.plt_units).magnitude
                yl = obj.plot_transform_c(obj.c)
                y_label = f"{obj.legend_left}"
            else:
                raise ValueError("plot_transform_c must be function")
        else:
            yl = (obj.c * model.c_unit).to(obj.plt_units).magnitude
            y_label = f"{obj.legend_left} [{obj.plt_units:~P}]"

    elif isinstance(obj, Signal):
        # use the same units as the associated flux
        yl = (obj.data.m * model.f_unit).to(obj.data.plt_units).magnitude
        y_label = f"{obj.n} [{obj.data.plt_units:~P}]"

    elif isinstance(obj, DataField):
        # time = (time * model.t_unit).to(model.d_unit).magnitude
        # yl = obj.y1_data
        # y_label = obj.y1_label
        if type(obj.y2_data) == str:
            ptype = 2
        else:
            ptype = 0

    else:  # sources, sinks, external data should not show up here
        raise ValueError(f"{obj.n} = {type(obj)}")

    # decide what to plot
    if ptype == 0:
        first_axis = True
        second_axis = True
    elif ptype == 1:
        first_axis = False
        second_axis = True
    elif ptype == 2:
        first_axis = True
        second_axis = False

    # start subplot
    ax1 = plt.subplot(rows, cols, fn)

    # set color index
    cn = 0
    col = f"C{cn}"

    if first_axis:
        if isinstance(obj, DataField):
            y_label = obj.y1_label
            if not isinstance(obj.y1_data[0], str):
                for i, d in enumerate(obj.y1_data):  # loop over datafield list
                    y1_legend = obj.y1_legend[i]
                    # print(f"label = {y1_legend}")
                    ln1 = ax1.plot(
                        obj.x1_data[i], obj.y1_data[i], color=col, label=y1_legend
                    )
                    cn = cn + 1
                    col = f"C{cn}"

                ax1.set_xlabel(f"{model.time_label} [{model.d_unit:~P}]")
                ax1.set_ylabel(y_label)
                # remove unnecessary frame species
                ax1.spines["top"].set_visible(False)
                # set_y_limits(ax1, obj)
                # plt.legend()

        else:
            # print(f"yl[2] = {yl[1:-2]}")
            ln1 = ax1.plot(time[1:-2], yl[1:-2], color=col, label=y_label)
            cn = cn + 1
            col = f"C{cn}"

            ax1.set_xlabel(f"{model.time_label} [{model.d_unit:~P}]")
            ax1.set_ylabel(y_label)
            # remove unnecessary frame species
            ax1.spines["top"].set_visible(False)
            # set_y_limits(ax1, obj)

    if second_axis:
        if isinstance(obj, DataField):
            y_label = obj.y2_label
            if not isinstance(obj.y2_data[0], str):
                if obj.common_y_scale == "yes":
                    for i, d in enumerate(obj.y2_data):  # loop over datafield list
                        y2_legend = obj.y2_legend[i]
                        ln1 = ax1.plot(
                            obj.x2_data[i], obj.y2_data[i], color=col, label=y2_legend
                        )
                        cn = cn + 1
                        col = f"C{cn}"
                        # set_y_limits(ax1, model)
                        # ax1.legend()
                        second_axis = False
                else:
                    ax2 = ax1.twinx()  # create a second y-axis
                    for i, d in enumerate(obj.y2_data):  # loop over datafield list
                        y2_legend = obj.y2_legend[i]
                        ln1 = ax2.plot(
                            obj.x2_data[i], obj.y2_data[i], color=col, label=y2_legend
                        )
                        cn = cn + 1
                        col = f"C{cn}"

                    ax2.set_ylabel(obj.y2_label)  # species object delta label
                    set_y_limits(ax2, model)
                    # remove unneeded frame
                    ax2.spines["top"].set_visible(False)
            else:
                second_axis = False

        elif isinstance(obj, Signal):
            # use the same units as the associated flux
            ax2 = ax1.twinx()  # create a second y-axis
            # plof right y-scale data
            ln2 = ax2.plot(
                time[1:-2], obj.data.d[1:-2], color=col, label=obj.legend_right
            )
            ax2.set_ylabel(obj.data.ld)  # species object delta label
            set_y_limits(ax2, model)
            ax2.spines["top"].set_visible(False)  # remove unnecessary frame speciess
        elif isinstance(obj, (Reservoir, GasReservoir)):
            ax2 = ax1.twinx()  # create a second y-axis
            # plof right y-scale data
            ln2 = ax2.plot(time[1:-2], yr[1:-2], color=col, label=obj.legend_right)
            ax2.set_ylabel(obj.ld)  # species object delta label
            set_y_limits(ax2, model)
            ax2.spines["top"].set_visible(False)  # remove unnecessary frame species
        elif isinstance(obj, Flux):
            # use the same units as the associated flux
            ax2 = ax1.twinx()  # create a second y-axis
            # plof right y-scale data
            ln2 = ax2.plot(time[1:-2], obj.d[1:-2], color=col, label=obj.legend_right)
            ax2.set_ylabel(obj.ld)  # species object delta label
            set_y_limits(ax2, model)
            ax2.spines["top"].set_visible(False)  # remove unnecessary frame speciess

    # adjust display properties for title and legend

    plt.rcParams["axes.titlepad"] = 20  # offset title upwards
    plt.rcParams["legend.facecolor"] = "0.8"  # show a gray background
    plt.rcParams["legend.edgecolor"] = "0.8"  # make frame the same color
    plt.rcParams["legend.framealpha"] = 0.4  # set transparency
    ax1.set_title(obj.full_name)

    for d in obj.led:  # loop over external data objects if present

        if isinstance(d.x[0], str):  # if string, something is off
            raise ValueError("No time axis in external data object {d.name}")
        if "y" in dir(d):  # mass or concentration data is present
            cn = cn + 1
            col = f"C{cn}"
            leg = f"{obj.lm} {d.legend}"
            ln3 = ax1.scatter(d.x[1:-2], d.y[1:-2], color=col, label=leg)
        if "z" in dir(d) and second_axis:  # isotope data is present
            cn = cn + 1
            col = f"C{cn}"
            leg = f"{d.legend}"
            ln3 = ax2.scatter(d.x, d.z, color=col, label=leg)

    # collect all labels and print them in one legend
    if first_axis:
        handler1, label1 = ax1.get_legend_handles_labels()
        plt.gca().spines["right"].set_visible(False)

    if second_axis:
        handler2, label2 = ax2.get_legend_handles_labels()

    if first_axis and second_axis:
        legend = ax2.legend(handler1 + handler2, label1 + label2, loc=0).set_zorder(6)
    elif first_axis:
        legend = ax1.legend(handler1, label1, loc=0).set_zorder(6)
    elif second_axis:
        legend = ax2.legend(handler2, label2, loc=0).set_zorder(6)
    else:
        raise TypeError("This should never happen!")


def is_name_in_list(n: str, l: list) -> bool:
    """Test if an object name is part of the object list"""

    r: bool = False
    for e in l:
        if e.full_name == n:
            r = True
    return r


def get_object_from_list(name: str, l: list) -> any:
    """Match a name to a list of objects. Return the object"""

    match: bool = False
    for o in l:
        if o.full_name == name:
            r = o
            match = True

    if match:
        return r
    else:
        raise ValueError(f"Object = {o.full_name} has no matching flux {name}")


def sort_by_type(l: list, t: list, m: str) -> list:
    """divide a list by type into new lists. This function will return a
    list and it is up to the calling code to unpack the list

    l is list with various object types
    t is a list which contains the object types used for sorting
    m is a string for the error function
    """

    # from numbers import Number

    lc = l.copy()
    rl = []

    for ot in t:  # loop over object types
        a = []
        for e in l:  # loop over list elements
            if isinstance(e, ot):
                a.append(e)  # add to temporary list
                lc.remove(e)  # remove this element

        rl.append(a)  # save the temporary list to rl

    # at this point, all elements of lc should have been processed
    # if not, lc contains element which are of a different type
    if len(lc) > 0:
        raise TypeError(m)

    return rl


def get_object_handle(res, M):
    """Test if we the key is a global reservoir handle
    or exists in the model namespace

    res: list, str, or reservoir handle
    M: Model handle
    """

    rlist: list = []

    if not isinstance(res, list):
        res = [res]

    for o in res:
        if o in M.dmo:  # is object known in global namespace
            rlist.append(M.dmo[o])
        elif o in M.__dict__:  # or does it exist in Model namespace
            rlist.append(getattr(M, o))
        else:
            raise ValueError(f"{o} is not known for model {M.name}")

    if len(rlist) == 1:
        rlist = rlist[0]

    return rlist


def split_key(k: str, M: any) -> Union[any, any, str]:
    """split the string k with letter 2, and test if optional
    id string is present

    """

    if "2" in k:
        source = k.split("2")[0]
        sinkandid = k.split("2")[1]
    else:
        raise ValueError("Name must follow 'Source2Sink' format")

    if "@" in sinkandid:
        sink = sinkandid.split("@")[0]
        cid = sinkandid.split("@")[1]
    else:
        sink = sinkandid
        cid = ""

    sink = get_object_handle(sink, M)
    source = get_object_handle(source, M)

    return (source, sink, cid)


def make_dict(keys: list, values: list) -> dict:
    """Create a dictionary from a list and value, or from
    two lists

    """
    if isinstance(values, list):
        if len(values) == len(keys):
            d: dict = dict(zip(keys, values))
        else:
            print(f"len values ={len(values)}, len keys ={len(keys)}")
            print(f"values = {values}")
            for k in keys:
                print(f"key = {k}")
            raise ValueError(f"key and value list must be of equal length")
    else:
        values: list = [values] * len(keys)
        d: dict = dict(zip(keys, values))

    return d


def get_typed_list(data: list) -> list:

    tl = List()
    for x in data:
        tl.append(x)
    return tl


def create_reservoirs(bn: dict, ic: dict, M: any, register: any = "None") -> dict:
    """boxes are defined by area and depth interval here we use an ordered
    dictionary to define the box geometries. The next column is temperature
    in deg C, followed by pressure in bar
    the geometry is [upper depth datum, lower depth datum, area percentage]

    bn = dictionary with box parameters
    bn: dict = {  # name: [[geometry], T, P]
                 "sb": {"g": [0, 200, 0.9], "T": 20, "P": 5},
                 "ib": {"g": [200, 1200, 1], "T": 10, "P": 100},
                }

    ic = dictionary with species default values. This is used to et up
         initial conditions. Here we use shortcut and use the same conditions
         in each box. If you need box specific initial conditions
         use the output of build_concentration_dicts as starting point

    ic: dict = { # species: concentration, Isotopes
                   PO4: [Q_("2.1 * umol/liter"), False],
                   DIC: [Q_("2.1 mmol/liter"), False],
                   ALK: [Q_("2.43 mmol/liter"), False],
               }

    M: Model object handle

    register reservoir groups in global name space (default), or with the
    provided object reference
    """

    from esbmtk import SeawaterConstants, ReservoirGroup, build_concentration_dicts
    from esbmtk import SourceGroup, SinkGroup, Q_

    # parse for sources and sinks, create these and remove them from the list

    # loop over reservoir names
    for k, v in bn.items():
        # test key format
        if M.name in k:
            k = k.split(".")[1]

        if "ty" in v:  # type is given
            if v["ty"] == "Source":
                SourceGroup(name=k, species=v["sp"], register=register)
            elif v["ty"] == "Sink":
                SinkGroup(name=k, species=v["sp"], register=register)
            else:
                raise ValueError("'ty' must be either Source or Sink")

        else:  # create reservoirs

            icd: dict = build_concentration_dicts(ic, k)
            # print(f"isotopes= {icd[k][1]}, delta = {icd[k][2]}")
            rg = ReservoirGroup(
                name=k,
                geometry=v["g"],
                concentration=icd[k][0],
                isotopes=icd[k][1],
                delta=icd[k][2],
                seawater_parameters={"temperature": v["T"], "pressure": v["P"]},
                register=register,
            )

    return icd


def build_concentration_dicts(cd: dict, bg: dict) -> dict:
    """Build a dict which can be used by create_reservoirs

    bg : dict where the box_names are dict keys.
    cd: dictionary with the following format:
        cd = {
             # species: [concentration, isotopes]
             PO4: [Q_("2.1 * umol/liter"), False],
             DIC: [Q_("2.1 mmol/liter"), False],
            }

    This function returns a new dict in the following format

    #  box_names: [concentrations, isotopes]
    d= {"bn": [{PO4: .., DIC: ..},{PO4:False, DIC:False}]}

    """

    if isinstance(bg, dict):
        box_names: list = bg.keys()
    elif isinstance(bg, list):
        box_names: list = bg
    elif isinstance(bg, str):
        box_names: list = [bg]
    else:
        raise ValueError("This should never happen")

    icd: dict = OrderedDict()
    td1: dict = {}  # temp dictionary
    td2: dict = {}  # temp dictionary
    td3: dict = {}  # temp dictionary

    # create the dicts for concentration and isotopes
    for k, v in cd.items():
        td1.update({k: v[0]})
        td2.update({k: v[1]})
        td3.update({k: v[2]})

    # box_names: list = bg.keys()
    for bn in box_names:  # loop over box names
        icd.update({bn: [td1, td2, td3]})

    return icd


def calc_volumes(bg: dict, M: any, h: any) -> list:
    """Calculate volume contained in a given depth interval
    bg is an ordered dictionary in the following format

    bg=  {
          "hb": (0.1, 0, 200),
          "sb": (0.9, 0, 200),
         }

    where the key must be a valid box name, the first entry of the list denoted
    the areal extent in percent, the second number is upper depth limit, and last
    number is the lower depth limit.

    M must be a model handle
    h is the hypsometry handle

    The function returns a list with the corresponding volumes

    """

    # from esbmtk import hypsometry

    v: list = []  # list of volumes

    for k, v in bg.items():
        a = v[0]
        u = v[1]
        l = v[2]

        v.append(h.volume(u, l) * a)

    return v


def get_longest_dict_entry(d: dict) -> int:
    """Get length of each item in the connection dict"""
    l_length = 0  # length of  longest list
    p_length = 0  # length of single parameter
    nl = 0  # number of lists
    ll = []

    # we need to cover the case where we have two lists of different length
    # this happens if we have a long list of tuples with matched references,
    # as well as a list of species
    for k, v in d.items():
        if isinstance(v, list):
            nl = nl + 1
            if len(v) > l_length:
                l_length = len(v)
                ll.append(l_length)

        else:
            p_length = 1

    if nl > 1:
        # if lists have different lengths
        if ll.count(l_length) != len(ll):
            raise ValueError("Mapping for multiple lists is not supported")

    if l_length > 0 and p_length == 0:
        case = 0  # Only lists present
    if l_length == 0 and p_length == 1:
        case = 1  #  Only parameters present
    if l_length > 0 and p_length == 1:
        case = 2  #  Lists and parameters present

    return case, l_length


def convert_to_lists(d: dict, l: int) -> dict:
    """expand mixed dict entries (i.e. list and single value) such
    that they are all lists of equal length

    """
    cd = d.copy()

    for k, v in cd.items():
        if not isinstance(v, list):
            p = []
            for i in range(l):
                p.append(v)
            d[k] = p

    return d


def get_sub_key(d: dict, i: int) -> dict:
    """take a dict which has where the value is a list, and return the
    key with the n-th value of that list

    """

    rd: dict = {}
    for k, v in d.items():
        rd[k] = v[i]

    return rd


def expand_dict(d: dict, mt: str = "1:1") -> int:
    """Determine dict structure

    in case we have mutiple connections with mutiple species, the
    default action is to map connections to species (t = '1:1'). If
    you rather want to create mutiple connections (one for each
    species) in each connection set t = '1:N'

    """
    # loop over dict entries
    # ck = connection key
    # cd = connection dict

    r: dict = {}  # the dict we will return

    for ck, cd in d.items():  # loop over connections

        # print(f"ck = {ck}")
        # print(f"cd = {cd}")

        rd: dict = {}  # temp dict
        nd: dict = {}  # temp dict
        case, length = get_longest_dict_entry(cd)

        # print(f"length = {length}")

        if isinstance(ck, tuple):
            # assume 1:1 mapping between tuple and connection parameters
            if mt == "1:1":

                # prep dictionaries
                if case == 0:
                    nd = cd
                    # print("case 0")
                elif case == 1:  # only parameters present. Expand for each tuple entry
                    length = len(ck)
                    nd = convert_to_lists(cd, length)
                    # print("case 1")
                elif case == 2:  # mixed list present, Expand list
                    # print(f"case 2, length = {length}")
                    nd = convert_to_lists(cd, length)
                    # print(nd)

                # for each connection group in the tuple
                if length != len(ck):
                    message = (
                        f"The number of connection properties ({length})\n"
                        f"does not match the number of connection groups ({len(ck)})\n"
                        f"did you intend to do a 1:N mapping?"
                    )
                    raise ValueError(message)

                # map property dicts to connection group names
                i = 0
                for t in ck:
                    rd[t] = get_sub_key(nd, i)
                    i = i + 1

            elif mt == "1:N":  # apply each species to each connection
                if case == 0:
                    nd = cd
                elif case == 1:  # only parameters present. Expand for each tuple entry
                    length = len(ck)
                    nd = convert_to_lists(cd, length)
                elif case == 2:  # mixed list present, Expand list
                    nd = convert_to_lists(cd, length)

                for t in ck:  # apply the entire nd dict to all connections
                    rd[t] = nd
            else:
                raise ValueError(f"{mt} is not defined. must be '1:1' or '1:N'")

        else:
            if case == 0:  # only lists present, case 3
                nd = cd
            elif case == 1:  # only parameters present
                nd = cd
            elif case == 2:  # list and parameters present case 4
                nd = convert_to_lists(cd, length)
            rd[ck] = nd

        # update the overall dict and move to the next entry
        r.update(rd)

    return r


def create_bulk_connections(ct: dict, M: any, mt: int = "1:1") -> None:
    """Create connections from a dictionary. The dict can have the following keys
    following format:

    mt = mapping type. See below for explanation

    # na: names, tuple or str. If lists, all list elements share the same properties
    # sp: species list or species
    # ty: type, str
    # ra: rate, Quantity
    # sc: scale, Number
    # re: reference, optional
    # al: alpha, optional
    # de: delta, optional
    # bp: bypass, see scale_with_flux
    # mx: True, optional defaults to False. If set, it will create forward
          and backward fluxes (i.e. mixing)

    There are 6 different cases how to specify connections

    Case 1 One connection, one set of parameters
           ct1 = {"sb2hb": {"ty": "scale", 'ra'....}}

    Case 2 One connection, one set of instructions, one subset with multiple parameters
           This will be expanded to create connections for each species
           ct2 = {"sb2hb": {"ty": "scale", "sp": ["a", "b"]}}

    Case 3 One connection complete set of multiple characters. Similar to case 2, but now
           all parameters are given explicitly
           ct3 = {"sb2hb": {"ty": ["scale", "scale"], "sp": ["a", "b"]}}

    Case 4 Multiple connections, one set of parameters. This will create
           identical connection for "sb2hb" and  "ib2db"
           ct4 = {("sb2hb", "ib2db"): {"ty": "scale", 'ra': ...}}

    Case 5 Multiple connections, one subset of multiple set of parameters. This wil
          create a connection for species 'a' in sb2hb and with species 'b' in ib2db
           ct5 = {("sb2hb", "ib2db"): {"ty": "scale", "sp": ["a", "b"]}}

    Case 6 Multiple connections, complete set of parameters of multiple parameters
           Same as case 5, but now all parameters are specified explicitly
           ct6 = {("sb2hb", "ib2db"): {"ty": ["scale", "scale"], "sp": ["a", "b"]}}


    The default interpretation for cases 5 and 6 is that each list
    entry corresponds to connection. However, sometimes we want to
    create multiple connections for multiple entries. In this case
    provide the mt='1:N' parameter which will create a connection for
    each species in each connection group. See the below example.

    It is easy to shoot yourself in the foot. It is best to try the above first with
    some simple examples, e.g.,

    from esbmtk import expand_dict
    ct2 = {"sb2hb": {"ty": "scale", "sp": ["a", "b"]}}

    It is best to use the show_dict function to verify that your input
    dictionary produces the correct results!

    """

    from esbmtk import create_connection, expand_dict

    # expand dictionary into a well formed dict where each connection
    # has a fully formed entry
    c_ct = expand_dict(ct, mt=mt)

    # loop over dict entries and create the respective connections
    for k, v in c_ct.items():
        if isinstance(k, tuple):  # loop over names in tuple
            for c in k:
                create_connection(c, v, M)
        elif isinstance(k, str):
            create_connection(k, v, M)
        else:
            raise ValueError(f"{connection} must be string or tuple")

    return ct


def create_connection(n: str, p: dict, M: any) -> None:
    """called by create_bulk_connections in order to create a connection
    group It is assumed that all rates are in liter/year or mol per
    year. This may not be what you want or need.

    n: a connection key e.g., sb2db@mix which will be
    interpreted as mixing a connection between sb and db and thus
    create connections in both directions

    p: a dictionary holding  the connection properties

    M: the model handle

    """

    from esbmtk import ConnectionGroup, Q_

    # get the reservoir handles by splitting the key
    if M.debug:
        print(f"key = {n}")
    source, sink, cid = split_key(n, M)

    # create default connections parameters and replace with values in
    # the parameter dict if present.
    los = list(p["sp"]) if isinstance(p["sp"], list) else [p["sp"]]
    typ = "None" if not "ty" in p else p["ty"]
    scale = 1 if not "sc" in p else p["sc"]
    rate = Q_("0 mol/a") if not "ra" in p else p["ra"]
    ref_flux = "None" if not "re" in p else p["re"]
    alpha = "None" if not "al" in p else p["al"]
    delta = "None" if not "de" in p else p["de"]
    mix = False if not "mx" in p else p["mx"]
    cid = f"{cid}_f" if mix else f"{cid}"
    bypass = "None" if not "bp" in p else p["bp"]
    species = "None" if not "sp" in p else p["sp"]

    if isinstance(scale, Q_):
        scale = scale.to("l/a").magnitude

    # if name in M.dmo: # group already exist in this case we nee to update the
    # connection group

    # Case one, no mixing, test if exist, otherwise create
    # Case 2, mixing: test if forward connection exists, if not create
    #                  test if backwards connection exists, or create

    if not mix:

        name = f"CG_{source.name}2{sink.name}"

        if M.debug:
            print(
                f"n = {n}, source: {source.full_name}, sink: {sink.full_name}, cid: {cid}, name = {name}"
            )

        update_or_create(
            name,
            source,
            sink,
            los,
            typ,
            scale,
            rate,
            ref_flux,
            alpha,
            delta,
            cid,
            M,
            bypass,
            species,
        )

    else:  # this is a connection with mixing
        # create forward connection
        name = f"CG_{source.name}2{sink.name}"

        update_or_create(
            name,
            source,
            sink,
            los,
            typ,
            scale,
            rate,
            ref_flux,
            alpha,
            delta,
            cid,
            M,
            bypass,
            species,
        )

        # create backwards connection
        name = f"CG_{sink.name}2{source.name}"
        cid = cid.replace("_f", "_b")

        update_or_create(
            name,
            sink,
            source,
            los,
            typ,
            scale,
            rate,
            ref_flux,
            alpha,
            delta,
            cid,
            M,
            bypass,
            species,
        )


def update_or_create(
    name,
    source,
    sink,
    los,
    typ,
    scale,
    rate,
    ref_flux,
    alpha,
    delta,
    cid,
    M,
    bypass,
    species,
):
    """Create or update connection"""

    from esbmtk import ConnectionGroup, Connect

    if M.register == "local":
        register = M
    else:
        register = "None"

    if M.debug:
        print(f"Looking up {name}")
    # update connection if already known
    if f"{name}" in M.lmo or f"{M.name}.{name}" in M.lmo:

        if M.register == "local":
            cg = getattr(M, name)
        else:
            cg = __builtins__[name]

        if M.debug:
            print(f"Connectiongroup is already known with full_name = {cg.full_name}")

        if isinstance(species, list):
            cg.update(
                name=name,
                source=source,
                sink=sink,
                ctype=make_dict(los, typ),
                scale=make_dict(los, scale),  # get rate from dictionary
                rate=make_dict(los, rate),
                ref_flux=make_dict(los, ref_flux),
                alpha=make_dict(los, alpha),
                delta=make_dict(los, delta),
                bypass=make_dict(los, bypass),
                register=register,
                id=cid,  # get id from dictionary
            )

        else:
            source = getattr(cg.source, f"{species.name}")
            sink = getattr(cg.sink, f"{species.name}")
            if M.debug:
                print(f"source = {source.full_name},  sink = {sink.full_name}")

            # #add new connection to the existing connection group
            g = Connect(
                # name=f"C_{id}",
                source=source,
                sink=sink,
                ctype=typ,
                scale=scale,  # get rate from dictionary
                rate=rate,
                ref_flux=ref_flux,
                alpha=alpha,
                delta=delta,
                bypass=bypass,
                register=cg,
                id=cid,  # get id from dictionary
            )
            if M.debug:
                print(f"Added {g.name} to {cg.full_name}")

    else:  # create new connection group
        cg = ConnectionGroup(
            # name=name,
            source=source,
            sink=sink,
            ctype=make_dict(los, typ),
            scale=make_dict(los, scale),  # get rate from dictionary
            rate=make_dict(los, rate),
            ref_flux=make_dict(los, ref_flux),
            alpha=make_dict(los, alpha),
            delta=make_dict(los, delta),
            bypass=make_dict(los, bypass),
            register=M,
            id=cid,  # get id from dictionary
        )
        if M.debug:
            print(f"created new connectiongroup with full_name = {cg.full_name}")


def get_name_only(o: any) -> any:
    """Test if item is an esbmtk type. If yes, extract the name"""

    from esbmtk import Flux, Reservoir, ReservoirGroup, Species
    from esbmtk import Sink, Source, SourceGroup, SinkGroup
    from esbmtk import Process, DataField, VirtualReservoir

    if isinstance(o, (Flux, Reservoir, ReservoirGroup, Species)):
        r = o.full_name
    else:
        r = o

    return r


def get_simple_list(l: list) -> list:
    """return a list which only has the full name
    rather than all the object properties

    """

    from esbmtk import Flux, Reservoir, Species

    r: list = []
    for e in l:
        r.append(get_name_only(e))

    return r


def show_dict(d: dict, mt: str = "1:1") -> None:
    """show dict entries in an organized manner"""

    from esbmtk import expand_dict, get_simple_list, get_name_only

    ct = expand_dict(d, mt)
    for ck, cv in ct.items():
        print(f"{ck}")

        for pk, pv in cv.items():
            if isinstance(pv, list):
                x = get_simple_list(pv)
            else:
                x = get_name_only(pv)
            print(f"     {pk} : {x}")


def find_matching_fluxes(l: list, filter_by: str, exclude: str) -> list:
    """Loop over all reservoir in l, and extract the names of all fluxes
    which match the filter string. Return the list of names (not objects!)

    """

    lof: set = set()

    for r in l:
        for f in r.lof:
            if filter_by in f.full_name and exclude not in f.full_name:
                lof.add(f)

    return list(lof)


def reverse_key(key: str) -> str:
    """reverse a connection key e.g., sb2db@POM becomes db2sb@POM"""

    # print(f"key = {key}")
    l = key.split("@")
    left = l[0]
    # right = l[1]
    rs = left.split("2")
    r1 = rs[0]
    r2 = rs[1]

    return f"{r2}2{r1}"


def get_connection_keys(
    s: set, fstr: str, nstr: str, inverse: bool, exclude: str
) -> list:
    """extract connection keys from set of flux names, replace fstr with
    nstr so that the key can be used in create_bulk_connnections()

    The optional inverse parameter, can be used where in cases where the
    flux direction needs to be reversed, i.e., the returned key will not read
    sb2db@POM, but db2s@POM

    E.g., if

    s = ( M4.CG_P_sb2P_ib.PO4.POP_F)
    fstr = "POP"
    nstr = "POM_DIC"

    M4.CG_P_sb2P_ib.PO4.POP_F will become

    P_sb2P_ib@POM_DIC

    """

    cl: list = []

    for n in s:
        # get connection and flux name
        l = n.full_name.split(".")
        cn = l[1][3:]  # get key without leadinf C_
        if inverse:
            cn = reverse_key(cn)
        cn.replace(fstr, nstr)
        cn = f"{cn}@{nstr}"
        cl.append(cn)

    return cl


def gen_dict_entries(M: any, **kwargs) -> tuple:
    """find all fluxes which contain the reference string, and create
    matching keys which contain the target string. The function will
    return two lists, which can be used to create a dict for the
    create_bulk_connnection function.

    E.g., to create a dict which will create new fluxes based on
    existing fluxes

    dk:list, fl:list = gen_dict_entries(ref_id = 'POP', target_id = 'POM')

    this will find all fluxes with the POP-id and put these into
    fl. It will also generate a list with suitable connections keys
    (dk) which contain the 'POM' id.

    The optional inverse parameter, can be used where in cases where the
    flux direction needs to be reversed, i.e., the returned key will not read
    sb2db@POM, but db2s@POM

    """

    from esbmtk import Model

    reference = kwargs["ref_id"]
    target = kwargs["target_id"]
    if "inverse" in kwargs:
        inverse = kwargs["inverse"]
    else:
        inverse = False

    if "exclude" in kwargs:
        exclude_str = kwargs["exclude"]
    else:
        exclude_str = "None"

    if isinstance(M, Model):
        flist: list = find_matching_fluxes(
            M.loc, filter_by=reference, exclude=exclude_str
        )
    elif isinstance(M, list):
        flist: list = find_matching_fluxes(
            M,
            filter_by=reference,
            exclude=exclude_str,
        )
    else:
        raise ValueError(f"gen_dict_entries: M must be list or Model, not {type(M)}")

    klist: list = get_connection_keys(flist, reference, target, inverse, exclude_str)

    return tuple(klist), flist


def build_ct_dict(d: dict, p: dict) -> dict:
    """build a connection dictionary from a dict containing connection
    keys, and a dict containing connection properties. This is most
    useful for connections which a characterized by a fixed rate but
    apply to many species. E.g., mixing fluxes in a complex model etc.

    """

    ct: dict = {}

    for k, v in d.items():
        td: dict = {}  # temp dict for scale
        td["sc"] = v
        td.update(p)
        ct[k] = td

    return ct


def get_string_between_brackets(s: str) -> str:
    """Parse string and extract substring between square brackets"""

    s = s.split("[")
    if len(s) < 2:
        raise ValueError(f"Column header {s} must include units in square brackets")

    s = s[1]

    s = s.split("]")

    if len(s) < 2:
        raise ValueError(f"Column header {s} must include units in square brackets")

    return s[0]


def check_for_quantity(kw) -> Q_:
    """check if keyword is quantity or string an convert as necessary

    kw = str or Q_

    """

    from esbmtk import Q_

    if isinstance(kw, str):
        kw = Q_(kw)
    elif isinstance(kw, Q_):
        pass
    else:
        raise ValueError(f"kw must be string or Quantity")

    return kw


def map_units(v: any, *args) -> float:
    """parse v to see if it is a string. if yes, map to quantity.
    parse v to see if it is a quantity, if yes, map to model units
    and extract magnitude, assign mangitude to return value
    if not, assign value to return value

    v : a keyword value number/string/quantity
    args: one or more quantities (units) see the Model class (e.g., f_unit)

    """

    from . import Q_

    m: float = 0
    match: bool = False

    # test if string, map to quantity if yes
    if isinstance(v, str):
        v = Q_(v)

    # test if we find a matching dimension, map if true
    if isinstance(v, Q_):
        for q in args:
            if v.dimensionality == q.dimensionality:
                m = v.to(q).magnitude
                match = True

        if not match:
            message = f"{v} is none of {print(*args)}"
            raise ValueError(message)

    else:  # no quantity, so it should be a number
        m = v

    if not isinstance(m, Number):
        raise ValueError(f"m is {type(m)}, must be float, v={v}. Something is fishy")

    return m


def add_carbonate_system_1(rgs: list):
    """Creates a new carbonate system virtual reservoir for each
    reservoir in rgs. Note that rgs must be a list of reservoir groups.

    Required keywords:
        rgs: list = []  of Reservoir Group objects

    These new virtual reservoirs are registered to their respective Reservoir
    as 'cs'.

    The respective data fields are available as rgs.r.cs.xxx where xxx stands
    for a given key key in the  vr_datafields dictionary (i.e., H, CA, etc.)

    """

    from esbmtk import ExternalCode, calc_carbonates_1

    # get object handle even if it defined in model namespace
    # rgs = get_object_handle(rgs)

    for rg in rgs:

        if rg.mo.register == "local":
            species = rg.mo.CO2
        else:
            species = __builtins__["CO2"]

        if hasattr(rg, "DIC") and hasattr(rg, "TA"):
            ExternalCode(
                name="cs",
                species=species,
                function=calc_carbonates_1,
                vr_datafields={
                    "H": rg.swc.hplus,  # 0
                    "CA": rg.swc.ca,  # 1
                    "HCO3": rg.swc.hco3,  # 2
                    "CO3": rg.swc.co3,  # 3
                    "CO2aq": rg.swc.co2,  # 4
                },
                function_input_data=List([rg.DIC.c, rg.TA.c]),
                function_params=List(
                    [
                        rg.swc.K1,  # 0
                        rg.swc.K2,  # 1
                        rg.swc.KW,  # 2
                        rg.swc.KB,  # 3
                        rg.swc.boron,  # 4
                        rg.swc.ca2,  # 5
                        rg.swc.Ksp,  # 6
                    ]
                ),
                register=rg,
            )
        else:
            raise AttributeError(f"{rg.full_name} must have a TA and DIC reservoir")


def add_carbonate_system_2(**kwargs) -> None:
    """Creates a new carbonate system virtual reservoir
    which will compute carbon species, saturation, compensation,
    and snowline depth, and compute the associated carbonate burial fluxes

    Required keywords:
        rgs: list of ReservoirGroup objects
        carbonate_export_fluxes: list of flux objects which mus match the
                                 list of ReservoirGroup objects.
        zsat_min = depth of the upper boundary of the deep box
        z0 = upper depth limit for carbonate burial calculations
             typically zsat_min

    Optional Parameters:

        zsat = initial saturation depth (m)
        zcc = initial carbon compensation depth (m)
        zsnow = initial snowline depth (m)
        zsat0 = characteristic depth (m)
        Ksp0 = solubility product of calcite at air-water interface (mol^2/kg^2)
        kc = heterogeneous rate constant/mass transfer coefficient for calcite dissolution (kg m^-2 yr^-1)
        Ca2 = calcium ion concentration (mol/kg)
        pc = characteristic pressure (atm)
        pg = seawater density multiplied by gravity due to acceleration (atm/m)
        I = dissolvable CaCO3 inventory
        co3 = CO3 concentration (mol/kg)
        Ksp = solubility product of calcite at in situ sea water conditions (mol^2/kg^2)

    """

    from esbmtk import carbonate_chemistry
    from esbmtk import ExternalCode, calc_carbonates_2

    # list of known keywords
    lkk: dict = {
        "rgs": list,
        "carbonate_export_fluxes": list,
        "AD": float,
        "zsat": int,
        "zsat_min": int,
        "zcc": int,
        "zsnow": int,
        "zsat0": int,
        "Ksp0": float,
        "kc": float,
        "Ca2": float,
        "pc": (float, int),
        "pg": (float, int),
        "I_caco3": (float, int),
        "alpha": float,
        "zmax": (float, int),
        "z0": (float, int),
        "Ksp": (float, int),
    }
    # provide a list of absolutely required keywords
    lrk: list[str] = ["rgs", "carbonate_export_fluxes", "zsat_min", "z0"]

    # we need the reference to the Model in order to set some
    # default values.

    reservoir = kwargs["rgs"][0]
    model = reservoir.mo
    # list of default values if none provided
    lod: dict = {
        "zsat": -3715,  # m
        "zcc": -4750,  # m
        "zsnow": -4750,  # m
        "zsat0": -5078,  # m
        "Ksp0": reservoir.swc.Ksp0,  # mol^2/kg^2
        "kc": 8.84 * 1000,  # m/yr converted to kg/(m^2 yr)
        "AD": model.hyp.area_dz(-200, -6000),
        "alpha": 0.6,  # 0.928771302395292, #0.75,
        "pg": 0.103,  # pressure in atm/m
        "pc": 511,  # characteristic pressure after Boudreau 2010
        "I_caco3": 529,  #  dissolveable CaCO3 in mol/m^2
        "zmax": -6000,  # max model depth
        "Ksp": reservoir.swc.Ksp,  # mol^2/kg^2
    }

    # make sure all mandatory keywords are present
    __checkkeys__(lrk, lkk, kwargs)

    # add default values for keys which were not specified
    kwargs = __addmissingdefaults__(lod, kwargs)

    # test that all keyword values are of the correct type
    __checktypes__(lkk, kwargs)

    # establish some shared parameters
    # depths_table = np.arange(0, 6001, 1)
    depths: NDArray = np.arange(0, 6002, 1, dtype=float)
    rgs = kwargs["rgs"]
    Ksp0 = kwargs["Ksp0"]
    ca2 = rgs[0].swc.ca2
    pg = kwargs["pg"]
    pc = kwargs["pc"]
    z0 = kwargs["z0"]
    Ksp = kwargs["Ksp"]

    # C saturation(z) after Boudreau 2010
    Csat_table: NDArray = (Ksp0 / ca2) * np.exp((depths * pg) / pc)
    area_table = model.hyp.get_lookup_table(0, -6002)  # area in m^2(z)
    area_dz_table = model.hyp.get_lookup_table_area_dz(0, -6002) * -1  # area'
    AD = model.hyp.area_dz(z0, -6000)  # Total Ocean Area

    for i, rg in enumerate(rgs):  # Setup the virtual reservoirs

        if rg.mo.register == "local":
            species = rg.mo.CO2
        else:
            species = __builtins__["CO2"]

        ExternalCode(
            name="cs",
            species=species,
            function=calc_carbonates_2,
            # datafield hold the results of the VR_no_set function
            # provide a default values which will be use to initialize
            # the respective datafield/
            vr_datafields={
                "H": rg.swc.hplus,  # 0 H+
                "CA": rg.swc.ca,  # 1 carbonate alkalinity
                "HCO3": rg.swc.hco3,  # 2 HCO3
                "CO3": rg.swc.co3,  # 3 CO3
                "CO2aq": rg.swc.co2,  # 4 CO2aq
                "zsat": kwargs["zsat"],  # 5 zsat
                "zcc": kwargs["zcc"],  # 6 zcc
                "zsnow": kwargs["zsnow"],  # 7 zsnow
                "Fburial": 0.0,  # 8 carbonate burial
                "Fburial12": 0.0,  # 9 carbonate burial 12C
                "diss": 0.0, # dissolution flux
                "Bm": 0.0,
            },
            function_input_data=List(
                [
                    rg.DIC.m,  # 0 DIC mass
                    rg.DIC.l,  # 1 DIC light isotope mass
                    rg.DIC.c,  # 2 DIC concentration
                    rg.TA.m,  # 3 TA mass
                    rg.TA.c,  # 4 TA concentration
                    kwargs["carbonate_export_fluxes"][i].fa,  # 5
                    area_table,  # 6
                    area_dz_table,  # 7
                    Csat_table,  # 8
                    rg.DIC.v,  # 9 reservoir volume
                ]
            ),
            function_params=List(
                [
                    rg.swc.K1,  # 0
                    rg.swc.K2,  # 1
                    rg.swc.KW,  # 2
                    rg.swc.KB,  # 3
                    rg.swc.boron,  # 4
                    Ksp0,  # 5
                    float(kwargs["kc"]),  # 6
                    float(rg.volume.to("liter").magnitude),  # 7
                    float(AD),  # 8
                    float(abs(kwargs["zsat0"])),  # 9
                    float(rg.swc.ca2),  # 10
                    rg.mo.dt,  # 11
                    float(kwargs["I_caco3"]),  # 12
                    float(kwargs["alpha"]),  # 13
                    float(abs(kwargs["zsat_min"])),  # 14
                    float(abs(kwargs["zmax"])),  # 15
                    float(abs(kwargs["z0"])),  # 16
                    Ksp,  # 17
                ]
            ),
            register=rg,
        )


def weathering_processes():

    from esbmtk import ExternalCode, calc_carbonates_1

    pass


def __find_flux__(reservoirs: list, full_name: str):
    """Helper function to find a Flux object based on its full_name in the reservoirs
    in the list of provided reservoirs.

    PRECONDITIONS: full_name must contain the full_name of the Flux

    Parameters:
        reservoirs: List containing all reservoirs
        full_name: str specifying the full name of the flux (boxes.flux_name)
    """
    needed_flux = None
    for res in reservoirs:
        for flux in res.lof:
            if flux.full_name == full_name:
                needed_flux = flux
                break
        if needed_flux != None:
            break
    if needed_flux == None:
        raise NameError(
            f"add_carbonate_system: Flux {full_name} cannot be found in any of the reservoirs in the Model!"
        )

    return needed_flux


def __checktypes__(av: Dict[any, any], pv: Dict[any, any]) -> None:
    """this method will use the the dict key in the user provided
    key value data (pv) to look up the allowed data type for this key in av

    av = dictinory with the allowed input keys and their type
    pv = dictionary with the user provided key-value data
    """

    k: any
    v: any

    # loop over provided keywords
    for k, v in pv.items():
        # check av if provided value v is of correct type
        if av[k] != any:
            # print(f"key = {k}, value  = {v}")
            if not isinstance(v, av[k]):
                # print(f"k={k}, v= {v}, av[k] = {av[k]}")
                raise TypeError(
                    f"{type(v)} is the wrong type for '{k}', should be '{av[k]}'"
                )


def __checkkeys__(lrk: list, lkk: list, kwargs: dict) -> None:
    """check if the mandatory keys are present

    lrk = list of required keywords
    lkk = list of all known keywords
    kwargs = dictionary with key-value pairs

    """

    k: str
    v: any
    # test if the required keywords are given
    for k in lrk:  # loop over required keywords
        if isinstance(k, list):  # If keyword is a list
            s: int = 0  # loop over allowed substitutions
            for e in k:  # test how many matches are in this list
                if e in kwargs:
                    # print(self.kwargs[e])
                    if not isinstance(e, (np.ndarray, np.float64, list)):
                        # print (type(self.kwargs[e]))
                        if kwargs[e] != "None":
                            s = s + 1
            if s > 1:  # if more than one match
                raise ValueError(f"You need to specify exactly one from this list: {k}")

        else:  # keyword is not in list
            if k not in kwargs:
                raise ValueError(f"You need to specify a value for {k}")

    tl: List[str] = []
    # create a list of known keywords
    for k, v in lkk.items():
        tl.append(k)

    # test if we know all keys
    for k, v in kwargs.items():
        if k not in lkk:
            raise ValueError(f"{k} is not a valid keyword. \n Try any of \n {tl}\n")


def __addmissingdefaults__(lod: dict, kwargs: dict) -> dict:
    """
    test if the keys in lod exist in kwargs, otherwise add them with the default values
    from lod

    """

    new: dict = {}
    if len(lod) > 0:
        for k, v in lod.items():
            if k not in kwargs:
                new.update({k: v})

    kwargs.update(new)
    return kwargs
